/*
 * Copyright (c) 2007-2009 Marko Kreen, Skype Technologies OÜ
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * @file
 *
 * File access utils.
 */
#ifndef _USUAL_FILEUTIL_H_
#define _USUAL_FILEUTIL_H_

#include <usual/base.h>

#include <stdio.h>

/** Info about mapped file */
struct MappedFile {
	int fd;
	unsigned len;
	void *ptr;
};

/** Signature for per-line callback */
typedef bool (*procline_cb)(void *arg, const char *line, ssize_t len);

/** Read file into memory */
void *load_file(const char *fn, size_t *len_p);

/** Loop over lines in file */
bool foreach_line(const char *fn, procline_cb proc_line, void *arg);

/** Get file size */
ssize_t file_size(const char *fn);

/** Map file into memory */
int map_file(struct MappedFile *m, const char *fname, int rw) _MUSTCHECK;

/** Unmap previously mapped file */
void unmap_file(struct MappedFile *m);

#if !defined(HAVE_GETLINE)
#define getline(a, b, c) compat_getline(a, b, c)

/**
 * Compat: Read line from file
 */
int getline(char **line_p, size_t *size_p, void *f);

#endif


#endif
